"""Define playing indicators for simple events.

This submodules provides several classes to add specific musical
playing techniques to :class:`mutwo.events.basic.SimpleEvent` objects.
They mostly derive from traditional Western playing techniques and their
notation. Unlike indicators of the :mod:`mutwo.parameters.notation_indicators`
module, playing indicators have an effect on the played music and aren't
merely specifications of representation. The proper way to handle
playing  indicators should be via a :class:`PlayingIndicatorCollection`
object that should be attached to the respective :class:`SimpleEvent`.
The collection contains all possible playing indicators which are defined
in this module. :class:`mutwo.events.music.NoteLike` contain by default
a playing indicator collection.

There are basically two different types of playing indicators:

1, Playing indicators which can only be on or off (for instance
``bartok_pizzicato``, ``prall`` or ``laissez_vibrer``). They have
a :attr:`is_active` attribute which can either be :obj:`True`
or :obj:`False`.

2. Playing indicators with one or more arguments (for instance
:class:`Tremolo` with :attr:`n_flags` or :class:`Arpeggio` with
:attr:`direction`). Their :attr:`is_active` attribute can't be
set by the user and get automatically initialised depending on
if all necessary attributes are defined (then active) or
if any of the necessary attributes is set to :obj:`None` (then
not active).

**Example:**

Set playing indicators of :class:`NoteLike`:

>>> from mutwo.events import music
>>> my_note = music.NoteLike('c', 1 / 4, 'mf')
>>> my_note.playing_indicators.articulation.name = "."  # add staccato
>>> my_chord = music.NoteLike('c e g', 1 / 2, 'f')
>>> my_chord.playing_indicators.arpeggio.direction= "up"  # add arpeggio
>>> my_chord.playing_indicators.laissez_vibrer = True  # and laissez_vibrer

Attach :class:`PlayingIndicatorCollection` to :class:`SimpleEvent`:

>>> from mutwo.events import basic
>>> from mutwo.parameters import playing_indicators
>>> my_simple_event = basic.SimpleEvent()
>>> my_simple_event.playing_indicators = playing_indicators.PlayingIndicatorCollection()
"""

import dataclasses
import typing

from mutwo.core import parameters

__all__ = (
    "Tremolo",
    "Articulation",
    "Arpeggio",
    "Pedal",
    "StringContactPoint",
    "Hairpin",
    "Ornamentation",
    "ArtificalHarmonic",
    "PreciseNaturalHarmonic",
    "Fermata",
    "PlayingIndicatorCollection",
)


@dataclasses.dataclass()
class Tremolo(parameters.abc.ImplicitPlayingIndicator):
    n_flags: typing.Optional[int] = None


@dataclasses.dataclass()
class Articulation(parameters.abc.ImplicitPlayingIndicator):
    name: typing.Optional[
        parameters.playing_indicators_constants.ARTICULATION_LITERAL
    ] = None


@dataclasses.dataclass()
class Arpeggio(parameters.abc.ImplicitPlayingIndicator):
    direction: typing.Optional[
        parameters.playing_indicators_constants.DIRECTION_LITERAL
    ] = None


@dataclasses.dataclass()
class Pedal(parameters.abc.ImplicitPlayingIndicator):
    pedal_type: typing.Optional[
        parameters.playing_indicators_constants.PEDAL_TYPE_LITERAL
    ] = None
    pedal_activity: typing.Optional[bool] = True


@dataclasses.dataclass()
class StringContactPoint(parameters.abc.ImplicitPlayingIndicator):
    contact_point: typing.Optional[
        parameters.playing_indicators_constants.CONTACT_POINT_LITERAL
    ] = None


@dataclasses.dataclass()
class Ornamentation(parameters.abc.ImplicitPlayingIndicator):
    direction: typing.Optional[
        parameters.playing_indicators_constants.DIRECTION_LITERAL
    ] = None
    n_times: int = 1


@dataclasses.dataclass()
class BendAfter(parameters.abc.ImplicitPlayingIndicator):
    # Content parameters
    bend_amount: typing.Optional[float] = None
    # Presentation parameters
    minimum_length: typing.Optional[float] = 3
    thickness: typing.Optional[float] = 3


@dataclasses.dataclass()
class ArtificalHarmonic(parameters.abc.ImplicitPlayingIndicator):
    n_semitones: typing.Optional[int] = None


@dataclasses.dataclass()
class PreciseNaturalHarmonic(parameters.abc.ImplicitPlayingIndicator):
    string_pitch: typing.Optional[parameters.pitches.WesternPitch] = None
    played_pitch: typing.Optional[parameters.pitches.WesternPitch] = None
    harmonic_note_head_style: bool = True
    parenthesize_lower_note_head: bool = False


@dataclasses.dataclass()
class Fermata(parameters.abc.ImplicitPlayingIndicator):
    fermata_type: typing.Optional[
        parameters.playing_indicators_constants.FERMATA_TYPE_LITERAL
    ] = None


@dataclasses.dataclass()
class Hairpin(parameters.abc.ImplicitPlayingIndicator):
    symbol: typing.Optional[
        parameters.playing_indicators_constants.HAIRPIN_SYMBOL_LITERAL
    ] = None


@dataclasses.dataclass()
class Trill(parameters.abc.ImplicitPlayingIndicator):
    pitch: typing.Optional[parameters.abc.Pitch] = None


@dataclasses.dataclass
class PlayingIndicatorCollection(
    parameters.abc.IndicatorCollection[parameters.abc.PlayingIndicator]
):
    # this is kind of redundant, but perhaps still better than without using
    # the `dataclasses` module
    articulation: Articulation = dataclasses.field(default_factory=Articulation)
    artifical_harmonic: ArtificalHarmonic = dataclasses.field(
        default_factory=ArtificalHarmonic
    )
    arpeggio: Arpeggio = dataclasses.field(default_factory=Arpeggio)
    bartok_pizzicato: parameters.abc.PlayingIndicator = dataclasses.field(
        default_factory=parameters.abc.ExplicitPlayingIndicator
    )
    bend_after: BendAfter = dataclasses.field(default_factory=BendAfter)
    breath_mark: parameters.abc.PlayingIndicator = dataclasses.field(
        default_factory=parameters.abc.ExplicitPlayingIndicator
    )
    duration_line_dashed: parameters.abc.PlayingIndicator = dataclasses.field(
        default_factory=parameters.abc.ExplicitPlayingIndicator
    )
    duration_line_triller: parameters.abc.PlayingIndicator = dataclasses.field(
        default_factory=parameters.abc.ExplicitPlayingIndicator
    )
    fermata: Fermata = dataclasses.field(default_factory=Fermata)
    glissando: parameters.abc.PlayingIndicator = dataclasses.field(
        default_factory=parameters.abc.ExplicitPlayingIndicator
    )
    hairpin: Hairpin = dataclasses.field(default_factory=Hairpin)
    natural_harmonic: parameters.abc.PlayingIndicator = dataclasses.field(
        default_factory=parameters.abc.ExplicitPlayingIndicator
    )
    laissez_vibrer: parameters.abc.PlayingIndicator = dataclasses.field(
        default_factory=parameters.abc.ExplicitPlayingIndicator
    )
    ornamentation: Ornamentation = dataclasses.field(default_factory=Ornamentation)
    pedal: Pedal = dataclasses.field(default_factory=Pedal)
    prall: parameters.abc.PlayingIndicator = dataclasses.field(
        default_factory=parameters.abc.ExplicitPlayingIndicator
    )
    precise_natural_harmonic: PreciseNaturalHarmonic = dataclasses.field(
        default_factory=PreciseNaturalHarmonic
    )
    string_contact_point: StringContactPoint = dataclasses.field(
        default_factory=StringContactPoint
    )
    tie: parameters.abc.PlayingIndicator = dataclasses.field(
        default_factory=parameters.abc.ExplicitPlayingIndicator
    )
    tremolo: Tremolo = dataclasses.field(default_factory=Tremolo)
    trill: Trill = dataclasses.field(default_factory=Trill)

    def __setattr__(self, parameter_name: str, value: bool):
        """Overriding default behaviour to allow syntactic sugar.

        This method allows syntax like:

            playing_indicator_collection.tie = True

        which is the same as

            playing_indicator_collection.tie.is_active = True

        Furthermore the methods makes sure that no property
        can actually be overridden.
        """

        try:
            playing_indicator = getattr(self, parameter_name)
        except AttributeError:
            playing_indicator = None
        if playing_indicator is not None:
            if isinstance(playing_indicator, parameters.abc.ExplicitPlayingIndicator):
                playing_indicator.is_active = bool(value)
            else:
                raise dataclasses.FrozenInstanceError(
                    "Can't override frozen property (playing indicator)"
                    f" '{playing_indicator}'!"
                )
        else:
            super().__setattr__(parameter_name, value)



"""This module adds several constants that are used for the different volume classes.

"""

MINIMUM_VELOCITY = 0
"""the lowest allowed midi velocity value"""

MAXIMUM_VELOCITY = 127
"""the highest allowed midi velocity value"""

# standard volume indicator
STANDARD_DYNAMIC_INDICATOR = tuple(
    "ppppp pppp ppp pp p mp mf f ff fff ffff fffff".split(" ")
)

DEFAULT_MINIMUM_DECIBEL_FOR_MIDI_VELOCITY_AND_STANDARD_DYNAMIC_INDICATOR: float = -40
"""Default value for ``minimum_decibel`` in
:class:`~mutwo.parameter.volumes.WesternVolume` and in
:method:`~mutwo.parameters.abc.Volume.decibel_to_midi_velocity`."""

DEFAULT_MAXIMUM_DECIBEL_FOR_MIDI_VELOCITY_AND_STANDARD_DYNAMIC_INDICATOR: float = 0
"""Default value for ``maximum_decibel`` in
:class:`~mutwo.parameter.volumes.WesternVolume` and in
:method:`~mutwo.parameters.abc.Volume.decibel_to_midi_velocity`."""


SPECIAL_DYNAMIC_INDICATOR_TO_STANDARD_DYNAMIC_INDICATOR_DICT = {
    "fp": "mf",
    "sf": "f",
    "sff": "ff",
    "sfz": "ff",
    "sp": "p",
    "spp": "pp",
    "rfz": "f",
}

DYNAMIC_INDICATOR_TUPLE = STANDARD_DYNAMIC_INDICATOR + tuple(
    SPECIAL_DYNAMIC_INDICATOR_TO_STANDARD_DYNAMIC_INDICATOR_DICT.keys()
)
"""all available dynamic indicator for :class:`WesternVolume`"""



"""Constants which are shared for all event classes in :mod:`mutwo.events.basic`."""

ROUND_DURATION_TO_N_DIGITS = 10
"""Set floating point precision for the duration property of all
:class:`~mutwo.events.abc.Event` classes in the :mod:`mutwo.basic`
module.

When returning the duration property all mentioned events will
round their actual duration if the duration type is float. This
behaviour has been added with version 0.28.1 to avoid floating
point rounding errors which could occur in all duration related
methods of the different event classes (as it can happen in
for instance the :func:`mutwo.events.abc.ComplexEvent.squash_in`
method or the :func:`mutwo.events.abc.Event.cut_off`
method)."""



"""Definition of global variables which are used all over mutwo.
"""

import fractions
import typing

Real = typing.Union[float, fractions.Fraction]
"""The main reason for this constant is a mypy issue with Pythons buildin
[numbers module](https://docs.python.org/3/library/numbers.html) which
is documented [here](https://github.com/python/mypy/issues/3186). Mypy
doesn't accept numbers abstract base classes. Until numbers will be
supported users have to define their own typing data for general number
classes. PEP 3141 recommends users to simply annotate arguments with
'float', but this wouldn't include `fractions.Fraction` which is often
necessary in musical contexts (as github user arseniiv also remarked)."""

DurationType = Real
"""Type variable to arguments and return values for `duration`.
This can be any real number (float, integer, fraction)."""

ParameterType = typing.Any
"""Type variable to assign to arguments and return values
which expect objects from the :mod:`mutwo.core.parameters` module,
but could actually be anything."""



"""Apply :class:`~mutwo.parameters.abc.PlayingIndicator` on :class:`~mutwo.events.abc.Event`.

"""

import abc
import copy
import itertools
import typing
import warnings

try:
    import quicktions as fractions  # type: ignore
except ImportError:
    import fractions  # type: ignore

from mutwo.core import converters
from mutwo.core import events
from mutwo.core import parameters
from mutwo.core.utilities import constants


__all__ = (
    "PlayingIndicatorConverter",
    "ArpeggioConverter",
    "ArticulationConverter",
    "PlayingIndicatorsConverter",
)


class PlayingIndicatorConverter(converters.abc.Converter):
    """Abstract base class to apply :class:`~mutwo.parameters.abc.PlayingIndicator` on a :class:`~mutwo.events.basic.SimpleEvent`.

    :param simple_event_to_playing_indicator_collection: Function to extract from a
        :class:`mutwo.events.basic.SimpleEvent` a
        :class:`mutwo.parameters.playing_indicators.PlayingIndicatorCollection`
        object. By default it asks the Event for its
        :attr:`~mutwo.events.music.NoteLike.playing_indicator_collection`
        attribute (because by default :class:`mutwo.events.music.NoteLike`
        objects are expected).
        When using different Event classes than :class:`~mutwo.events.music.NoteLike`
        with a different name for their playing_indicator_collection property, this argument
        should be overridden. If the
        function call raises an :obj:`AttributeError` (e.g. if no playing indicator
        collection can be extracted), mutwo will build a playing indicator collection
        from :const:`~mutwo.events.music_constants.DEFAULT_PLAYING_INDICATORS_COLLECTION_CLASS`.
    :type simple_event_to_playing_indicator_collection: typing.Callable[[events.basic.SimpleEvent], parameters.playing_indicators.PlayingIndicatorCollection], optional
    """

    def __init__(
        self,
        simple_event_to_playing_indicator_collection: typing.Callable[
            [events.basic.SimpleEvent],
            parameters.playing_indicators.PlayingIndicatorCollection,
        ] = lambda simple_event: simple_event.playing_indicator_collection,  # type: ignore
    ):
        self._simple_event_to_playing_indicator_collection = (
            simple_event_to_playing_indicator_collection
        )

    @abc.abstractmethod
    def _apply_playing_indicator(
        self,
        simple_event_to_convert: events.basic.SimpleEvent,
        playing_indicator_collection: parameters.playing_indicators.PlayingIndicatorCollection,
    ) -> events.basic.SequentialEvent[events.basic.SimpleEvent]:
        raise NotImplementedError()

    def convert(
        self, simple_event_to_convert: events.basic.SimpleEvent
    ) -> events.basic.SequentialEvent[events.basic.SimpleEvent]:
        """Apply PlayingIndicator on simple_event.

        :param simple_event_to_convert: The event which shall be converted.
        :type simple_event_to_convert: events.basic.SimpleEvent
        """

        simple_event_to_playing_indicator_collection = (
            self._simple_event_to_playing_indicator_collection
        )

        try:
            playing_indicator_collection = simple_event_to_playing_indicator_collection(
                simple_event_to_convert
            )
        except AttributeError:
            playing_indicator_collection = (
                events.music_constants.DEFAULT_PLAYING_INDICATORS_COLLECTION_CLASS()
            )

        return self._apply_playing_indicator(
            simple_event_to_convert, playing_indicator_collection
        )


class ArpeggioConverter(PlayingIndicatorConverter):
    """Apply arpeggio on :class:`~mutwo.events.basic.SimpleEvent`.

    :param duration_for_each_attack: Set how long each attack of the
        Arpeggio lasts. Default to 0.1.
    :type duration_for_each_attack: constants.DurationType
    :param simple_event_to_pitches: Function to extract from a
        :class:`mutwo.events.basic.SimpleEvent` a tuple that contains pitch objects
        (objects that inherit from :class:`mutwo.parameters.abc.Pitch`).
        By default it asks the Event for its
        :attr:`~mutwo.events.music.NoteLike.pitch_list` attribute
        (because by default :class:`mutwo.events.music.NoteLike` objects are expected).
        When using different Event classes than :class:`~mutwo.events.music.NoteLike`
        with a different name for their pitch property, this argument
        should be overridden.
        If the function call raises an :obj:`AttributeError` (e.g. if no pitch can be
        extracted), mutwo will assume an event without any pitches.
    :type simple_event_to_pitches: typing.Callable[[events.basic.SimpleEvent], parameters.abc.Pitch], optional
    :param simple_event_to_playing_indicator_collection: Function to extract from a
        :class:`mutwo.events.basic.SimpleEvent` a
        :class:`mutwo.parameters.playing_indicators.PlayingIndicatorCollection`
        object. By default it asks the Event for its
        :attr:`~mutwo.events.music.NoteLike.playing_indicator_collection`
        attribute (because by default :class:`mutwo.events.music.NoteLike`
        objects are expected).
        When using different Event classes than :class:`~mutwo.events.music.NoteLike`
        with a different name for their playing_indicator_collection property, this argument
        should be overridden. If the
        function call raises an :obj:`AttributeError` (e.g. if no playing indicator
        collection can be extracted), mutwo will build a playing indicator collection
        from :const:`~mutwo.events.music_constants.DEFAULT_PLAYING_INDICATORS_COLLECTION_CLASS`.
    :type simple_event_to_playing_indicator_collection: typing.Callable[[events.basic.SimpleEvent], parameters.playing_indicators.PlayingIndicatorCollection,], optional
    :param set_pitch_list_for_simple_event: Function which assigns
        a list of :class:`~mutwo.parameters.abc.Pitch` objects to a
        :class:`~mutwo.events.basic.SimpleEvent`. By default the
        function assigns the passed pitches to the
        :attr:`~mutwo.events.music.NoteLike.pitch_list` attribute
        (because by default :class:`mutwo.events.music.NoteLike` objects
        are expected).
    :type set_pitch_list_for_simple_event: typing.Callable[[events.basic.SimpleEvent, list[parameters.abc.Pitch]], None]
    """

    def __init__(
        self,
        duration_for_each_attack: constants.DurationType = 0.1,
        simple_event_to_pitches: typing.Callable[
            [events.basic.SimpleEvent], list[parameters.abc.Pitch]
        ] = lambda simple_event: simple_event.pitch_list,  # type: ignore
        simple_event_to_playing_indicator_collection: typing.Callable[
            [events.basic.SimpleEvent],
            parameters.playing_indicators.PlayingIndicatorCollection,
        ] = lambda simple_event: simple_event.playing_indicator_collection,  # type: ignore
        set_pitch_list_for_simple_event: typing.Callable[
            [events.basic.SimpleEvent, list[parameters.abc.Pitch]], None
        ] = lambda simple_event, pitch_list: simple_event.set_parameter(  # type: ignore
            "pitch_list", pitch_list, set_unassigned_parameter=True
        ),
    ):
        super().__init__(
            simple_event_to_playing_indicator_collection=simple_event_to_playing_indicator_collection
        )
        self._duration_for_each_attack = duration_for_each_attack
        self._simple_event_to_pitches = simple_event_to_pitches
        self._set_pitch_list_for_simple_event = set_pitch_list_for_simple_event

    def _apply_arpeggio(
        self,
        simple_event_to_convert: events.basic.SimpleEvent,
        arpeggio: parameters.playing_indicators.Arpeggio,
    ) -> events.basic.SequentialEvent[events.basic.SimpleEvent]:
        try:
            pitch_list = list(self._simple_event_to_pitches(simple_event_to_convert))
        except AttributeError:
            pitch_list = []

        # sort pitches according to Arpeggio direction
        pitch_list.sort(reverse=arpeggio.direction != "up")

        converted_event: events.basic.SequentialEvent[
            events.basic.SimpleEvent
        ] = events.basic.SequentialEvent(
            [copy.copy(simple_event_to_convert) for _ in pitch_list]
        )

        # apply pitches on events
        for nth_event, pitch in enumerate(pitch_list):
            self._set_pitch_list_for_simple_event(converted_event[nth_event], [pitch])

        # set correct duration for each event
        n_events = len(converted_event)
        duration_of_each_attack = self._duration_for_each_attack
        if n_events * duration_of_each_attack > simple_event_to_convert.duration:
            duration_of_each_attack = simple_event_to_convert.duration / n_events

        for nth_event in range(n_events - 1):
            converted_event[nth_event].duration = duration_of_each_attack

        converted_event[-1].duration -= (
            converted_event.duration - simple_event_to_convert.duration
        )

        return converted_event

    def _apply_playing_indicator(
        self,
        simple_event_to_convert: events.basic.SimpleEvent,
        playing_indicator_collection: parameters.playing_indicators.PlayingIndicatorCollection,
    ) -> events.basic.SequentialEvent[events.basic.SimpleEvent]:
        try:
            arpeggio = playing_indicator_collection.arpeggio
        except AttributeError:
            arpeggio = parameters.playing_indicators.Arpeggio()

        if arpeggio.is_active:
            return self._apply_arpeggio(simple_event_to_convert, arpeggio)
        else:
            return events.basic.SequentialEvent([copy.copy(simple_event_to_convert)])


class StacattoConverter(PlayingIndicatorConverter):
    """Apply staccato on :class:`~mutwo.events.basic.SimpleEvent`.

    :param factor:
    :param allowed_articulation_name_sequence:
    :param simple_event_to_playing_indicator_collection: Function to extract from a
        :class:`mutwo.events.basic.SimpleEvent` a
        :class:`mutwo.parameters.playing_indicators.PlayingIndicatorCollection`
        object. By default it asks the Event for its
        :attr:`~mutwo.events.music.NoteLike.playing_indicator_collection`
        attribute (because by default :class:`mutwo.events.music.NoteLike`
        objects are expected).
        When using different Event classes than :class:`~mutwo.events.music.NoteLike`
        with a different name for their playing_indicator_collection property, this argument
        should be overridden. If the
        function call raises an :obj:`AttributeError` (e.g. if no playing indicator
        collection can be extracted), mutwo will build a playing indicator collection
        from :const:`~mutwo.events.music_constants.DEFAULT_PLAYING_INDICATORS_COLLECTION_CLASS`.
    :type simple_event_to_playing_indicator_collection: typing.Callable[[events.basic.SimpleEvent], parameters.playing_indicators.PlayingIndicatorCollection,], optional
    """

    def __init__(
        self,
        factor: float = 0.5,
        allowed_articulation_name_sequence: typing.Sequence[str] = ("staccato", "."),
        simple_event_to_playing_indicator_collection: typing.Callable[
            [events.basic.SimpleEvent],
            parameters.playing_indicators.PlayingIndicatorCollection,
        ] = lambda simple_event: simple_event.playing_indicator_collection,  # type: ignore
    ):
        self._allowed_articulation_name_sequence = allowed_articulation_name_sequence
        self._factor = factor
        super().__init__(simple_event_to_playing_indicator_collection)

    def _apply_stacatto(
        self,
        simple_event_to_convert: events.basic.SimpleEvent,
    ) -> events.basic.SequentialEvent[events.basic.SimpleEvent]:
        duration = simple_event_to_convert.duration * self._factor
        sequential_event = events.basic.SequentialEvent(
            [
                simple_event_to_convert.set_parameter(
                    "duration", duration, mutate=False
                ),
                events.basic.SimpleEvent(duration),
            ]
        )
        return sequential_event

    def _apply_playing_indicator(
        self,
        simple_event_to_convert: events.basic.SimpleEvent,
        playing_indicator_collection: parameters.playing_indicators.PlayingIndicatorCollection,
    ) -> events.basic.SequentialEvent[events.basic.SimpleEvent]:
        try:
            articulation = playing_indicator_collection.articulation
        except AttributeError:
            articulation = parameters.playing_indicators.Articulation()

        if (
            articulation.is_active
            and articulation.name in self._allowed_articulation_name_sequence
        ):
            return self._apply_stacatto(simple_event_to_convert)
        else:
            return events.basic.SequentialEvent([copy.copy(simple_event_to_convert)])


class ArticulationConverter(PlayingIndicatorConverter):
    """Apply articulation on :class:`~mutwo.events.basic.SimpleEvent`.

    :param articulation_name_tuple_to_playing_indicator_converter:
    :type articulation_name_tuple_to_playing_indicator_converter: dict[tuple[str, ...], PlayingIndicatorConverter]
    :param simple_event_to_playing_indicator_collection: Function to extract from a
        :class:`mutwo.events.basic.SimpleEvent` a
        :class:`mutwo.parameters.playing_indicators.PlayingIndicatorCollection`
        object. By default it asks the Event for its
        :attr:`~mutwo.events.music.NoteLike.playing_indicator_collection`
        attribute (because by default :class:`mutwo.events.music.NoteLike`
        objects are expected).
        When using different Event classes than :class:`~mutwo.events.music.NoteLike`
        with a different name for their playing_indicator_collection property, this argument
        should be overridden. If the
        function call raises an :obj:`AttributeError` (e.g. if no playing indicator
        collection can be extracted), mutwo will build a playing indicator collection
        from :const:`~mutwo.events.music_constants.DEFAULT_PLAYING_INDICATORS_COLLECTION_CLASS`.
    :type simple_event_to_playing_indicator_collection: typing.Callable[[events.basic.SimpleEvent], parameters.playing_indicators.PlayingIndicatorCollection,], optional
    """

    def __init__(
        self,
        articulation_name_tuple_to_playing_indicator_converter: dict[
            tuple[str, ...], PlayingIndicatorConverter
        ] = {("staccato", "."): StacattoConverter()},
        simple_event_to_playing_indicator_collection: typing.Callable[
            [events.basic.SimpleEvent],
            parameters.playing_indicators.PlayingIndicatorCollection,
        ] = lambda simple_event: simple_event.playing_indicator_collection,  # type: ignore
    ):
        articulation_name_to_playing_indicator_converter = {}
        for (
            articulation_name_tuple,
            playing_indicator_converter,
        ) in articulation_name_tuple_to_playing_indicator_converter.items():
            for articulation_name in articulation_name_tuple:
                try:
                    assert (
                        articulation_name
                        not in articulation_name_to_playing_indicator_converter
                    )
                except AssertionError:
                    message = "Found two playing indicator converter mappings for "
                    message += f"articulation name '{articulation_name}'! "
                    message += "Mutwo will use the playing indicator converter "
                    message += f"'{playing_indicator_converter}'."
                    warnings.warn(message)
                articulation_name_to_playing_indicator_converter.update(
                    {articulation_name: playing_indicator_converter}
                )

        self._articulation_name_to_playing_indicator_converter = (
            articulation_name_to_playing_indicator_converter
        )
        super().__init__(simple_event_to_playing_indicator_collection)

    def _apply_playing_indicator(
        self,
        simple_event_to_convert: events.basic.SimpleEvent,
        playing_indicator_collection: parameters.playing_indicators.PlayingIndicatorCollection,
    ) -> events.basic.SequentialEvent[events.basic.SimpleEvent]:
        try:
            articulation = playing_indicator_collection.articulation
        except AttributeError:
            articulation = parameters.playing_indicators.Articulation()

        if (
            articulation.is_active
            and articulation.name
            in self._articulation_name_to_playing_indicator_converter
        ):
            return self._articulation_name_to_playing_indicator_converter[
                articulation.name
            ].convert(simple_event_to_convert)
        else:
            return events.basic.SequentialEvent([copy.copy(simple_event_to_convert)])


class TrillConverter(PlayingIndicatorConverter):
    """Apply trill on :class:`~mutwo.events.basic.SimpleEvent`.

    :param trill_size:
    :type trill_size: constants.DurationType
    :param simple_event_to_pitch_list: Function to extract from a
        :class:`mutwo.events.basic.SimpleEvent` a tuple that contains pitch objects
        (objects that inherit from :class:`mutwo.parameters.abc.Pitch`).
        By default it asks the Event for its
        :attr:`~mutwo.events.music.NoteLike.pitch_list` attribute
        (because by default :class:`mutwo.events.music.NoteLike` objects are expected).
        When using different Event classes than :class:`~mutwo.events.music.NoteLike`
        with a different name for their pitch property, this argument
        should be overridden.
        If the function call raises an :obj:`AttributeError` (e.g. if no pitch can be
        extracted), mutwo will assume an event without any pitches.
    :type simple_event_to_pitch_list: typing.Callable[[events.basic.SimpleEvent], parameters.abc.Pitch], optional
    :param simple_event_to_playing_indicator_collection: Function to extract from a
        :class:`mutwo.events.basic.SimpleEvent` a
        :class:`mutwo.parameters.playing_indicators.PlayingIndicatorCollection`
        object. By default it asks the Event for its
        :attr:`~mutwo.events.music.NoteLike.playing_indicator_collection`
        attribute (because by default :class:`mutwo.events.music.NoteLike`
        objects are expected).
        When using different Event classes than :class:`~mutwo.events.music.NoteLike`
        with a different name for their playing_indicator_collection property, this argument
        should be overridden. If the
        function call raises an :obj:`AttributeError` (e.g. if no playing indicator
        collection can be extracted), mutwo will build a playing indicator collection
        from :const:`~mutwo.events.music_constants.DEFAULT_PLAYING_INDICATORS_COLLECTION_CLASS`.
    :type simple_event_to_playing_indicator_collection: typing.Callable[[events.basic.SimpleEvent], parameters.playing_indicators.PlayingIndicatorCollection,], optional
    """

    def __init__(
        self,
        trill_size: constants.DurationType = fractions.Fraction(1, 16),
        simple_event_to_pitch_list: typing.Callable[
            [events.basic.SimpleEvent], list[parameters.abc.Pitch]
        ] = lambda simple_event: simple_event.pitch_list,  # type: ignore
        simple_event_to_playing_indicator_collection: typing.Callable[
            [events.basic.SimpleEvent],
            parameters.playing_indicators.PlayingIndicatorCollection,
        ] = lambda simple_event: simple_event.playing_indicator_collection,  # type: ignore
    ):
        self._trill_size = trill_size
        self._simple_event_to_pitch_list = simple_event_to_pitch_list
        super().__init__(simple_event_to_playing_indicator_collection)

    def _apply_trill(
        self,
        simple_event_to_convert: events.basic.SimpleEvent,
        trill: parameters.playing_indicators.Trill,
        pitch_list: list[parameters.abc.Pitch],
    ) -> events.basic.SequentialEvent[events.basic.SimpleEvent]:
        n_trill_items = simple_event_to_convert.duration // self._trill_size
        remaining = simple_event_to_convert.duration - (
            n_trill_items * self._trill_size
        )
        sequential_event = events.basic.SequentialEvent([])
        pitch_cycle = itertools.cycle((pitch_list, trill.pitch))
        for _ in range(int(n_trill_items)):
            simple_event = simple_event_to_convert.set_parameter(
                "duration", self._trill_size, mutate=False
            ).set_parameter("pitch_list", next(pitch_cycle))
            sequential_event.append(simple_event)
        sequential_event[-1].duration += remaining
        return sequential_event

    def _apply_playing_indicator(
        self,
        simple_event_to_convert: events.basic.SimpleEvent,
        playing_indicator_collection: parameters.playing_indicators.PlayingIndicatorCollection,
    ) -> events.basic.SequentialEvent[events.basic.SimpleEvent]:
        try:
            trill = playing_indicator_collection.trill
        except AttributeError:
            trill = parameters.playing_indicators.Trill()

        simple_event_to_pitch_list = self._simple_event_to_pitch_list
        try:
            pitch_list = simple_event_to_pitch_list(simple_event_to_convert)
        except AttributeError:
            pitch_list = []

        if (
            trill.is_active
            and simple_event_to_convert.duration > self._trill_size * 2
            and pitch_list
        ):
            return self._apply_trill(simple_event_to_convert, trill, pitch_list)
        else:
            return events.basic.SequentialEvent([copy.copy(simple_event_to_convert)])


class PlayingIndicatorsConverter(converters.abc.EventConverter):
    """Apply :class:`~mutwo.parameters.abc.PlayingIndicator` on any :class:`~mutwo.events.abc.Event`.

    :param playing_indicator_converters: A sequence of :class:`PlayingIndicatorConverter` which shall
        be applied on each :class:`~mutwo.events.basic.SimpleEvent`.
    :type playing_indicator_converters: typing.Sequence[PlayingIndicatorConverter]
    """

    def __init__(
        self,
        playing_indicator_converters: typing.Sequence[PlayingIndicatorConverter],
    ):
        self._playing_indicator_converters = playing_indicator_converters

    def _convert_simple_event(
        self,
        event_to_convert: events.basic.SimpleEvent,
        _: constants.DurationType,
    ) -> events.basic.SequentialEvent[events.basic.SimpleEvent]:
        """Convert instance of :class:`mutwo.events.basic.SimpleEvent`."""
        converted_event = [event_to_convert]

        for playing_indicator_converter in self._playing_indicator_converters:
            new_converted_event: list[events.basic.SimpleEvent] = []
            for simple_event in converted_event:
                converted_simple_event = playing_indicator_converter.convert(
                    simple_event
                )
                new_converted_event.extend(converted_simple_event)

            converted_event = new_converted_event

        return events.basic.SequentialEvent(converted_event)

    def convert(self, event_to_convert: events.abc.Event) -> events.abc.Event:
        return self._convert_event(event_to_convert, 0)



"""Define notation indicators for simple events.

This submodules provides several classes to express notation
specifications for :class:`mutwo.events.basic.SimpleEvent` objects.
They mostly derive from traditional Western notation.
Unlike indicators of the :mod:`mutwo.parameters.notation_indicators`
module, notation indicators shouldn't have an effect on the played music
and are merely specifications of representation. The proper way to handle
notation indicators should be via a :class:`NotationIndicatorCollection`
object that should be attached to the respective :class:`SimpleEvent`.
The collection contains all possible notation indicators which are defined
in this module. :class:`mutwo.events.music.NoteLike` contain by default
a notation indicator collection.

Notation indicators have one or more arguments. Their :attr:`is_active`
attribute can't be set by the user and get automatically initialised depending
on if all necessary attributes are defined (then active) or if any of the
necessary attributes is set to :obj:`None` (then not active).

**Example:**

Set notation indicators of :class:`NoteLike`:

>>> from mutwo.events import music
>>> my_note = music.NoteLike('c', 1 / 4, 'mf')
>>> my_note.notation_indicators.margin_markup.content = "Violin"
"""

import dataclasses
import typing

from mutwo.core import parameters


__all__ = (
    "BarLine",
    "Clef",
    "Ottava",
    "MarginMarkup",
    "Markup",
    "RehearsalMark",
    "NotationIndicatorCollection",
)


@dataclasses.dataclass()
class BarLine(parameters.abc.NotationIndicator):
    abbreviation: typing.Optional[
        str
    ] = None  # TODO(for future usage add typing.Literal)


@dataclasses.dataclass()
class Clef(parameters.abc.NotationIndicator):
    name: typing.Optional[str] = None  # TODO(for future usage add typing.Literal)


@dataclasses.dataclass()
class Ottava(parameters.abc.NotationIndicator):
    n_octaves: typing.Optional[int] = 0


@dataclasses.dataclass()
class MarginMarkup(parameters.abc.NotationIndicator):
    content: typing.Optional[str] = None
    context: typing.Optional[str] = "Staff"  # TODO(for future usage add typing.Literal)


@dataclasses.dataclass()
class Markup(parameters.abc.NotationIndicator):
    content: typing.Optional[str] = None
    direction: typing.Optional[str] = None


@dataclasses.dataclass()
class RehearsalMark(parameters.abc.NotationIndicator):
    markup: typing.Optional[str] = None


@dataclasses.dataclass
class NotationIndicatorCollection(
    parameters.abc.IndicatorCollection[parameters.abc.NotationIndicator]
):
    bar_line: BarLine = dataclasses.field(default_factory=BarLine)
    clef: Clef = dataclasses.field(default_factory=Clef)
    ottava: Ottava = dataclasses.field(default_factory=Ottava)
    margin_markup: MarginMarkup = dataclasses.field(default_factory=MarginMarkup)
    markup: Markup = dataclasses.field(default_factory=Markup)
    rehearsal_mark: RehearsalMark = dataclasses.field(default_factory=RehearsalMark)

    def __setattr__(self, parameter_name: str, value: bool):
        try:
            notation_indicator = getattr(self, parameter_name)
        except AttributeError:
            notation_indicator = None
        if notation_indicator is not None:
            message = f"Can't override frozen property (notation indicator) '{notation_indicator}'!"
            raise dataclasses.FrozenInstanceError(message)
        else:
            super().__setattr__(parameter_name, value)



"""Defining the public API for any converter class."""

import abc
import typing

from mutwo.core import events
from mutwo.core.utilities import constants

__all__ = ("Converter", "EventConverter", "SymmetricalEventConverter")


class Converter(abc.ABC):
    """Abstract base class for all Converter classes.

    Converter classes are defined as classes that convert data between
    two different encodings. Their only public method (besides initialisation)
    should be a `convert` method. The first argument of the convert method
    should be the data to convert.
    """

    @abc.abstractmethod
    def convert(
        self, event_or_parameter_or_file_to_convert: typing.Any, *args, **kwargs
    ) -> typing.Any:
        raise NotImplementedError


class EventConverter(Converter):
    """Abstract base class for Converter which handle mutwo events.

    This class helps building new classes which convert mutwo events
    with few general private methods (and without adding any new public
    method). Converting mutwo event often involves the same pattern:
    due to the nested structure of an Event, the converter has
    to iterate through the different layers until it reaches leaves
    (any class that inherits from :class:`mutwo.events.basic.SimpleEvent`).
    This common iteration process and the different time treatment
    between :class:`mutwo.events.basic.SequentialEvent` and
    :class:`mutwo.events.basic.SimultaneousEvent` are implemented in
    :class:`EventConverter`.  For writing a new EventConverter class,
    one only has to override the abstract method :func:`_convert_simple_event`
    and the abstract method :func:`convert` (where one will perhaps call
    :func:`_convert_event`.).

    **Example:**

    The following example defines a dummy class for demonstrating how
    to use EventConverter.

    >>> from mutwo.converters import abc
    >>> class DurationPrintConverter(abc.EventConverter):
    >>>     def _convert_simple_event(self, event_to_convert, absolute_entry_delay):
    >>>         return "{}: {}".format(absolute_entry_delay, event_to_convert.duration),
    >>>     def convert(self, event_to_convert):
    >>>         data_per_event = self._convert_event(event_to_convert, 0)
    >>>         [print(data) for data in data_per_event]
    >>> # now test with random event
    >>> import random
    >>> from mutwo.events import basic
    >>> random.seed(100)
    >>> random_event = basic.SimultaneousEvent(
    >>>     [
    >>>         basic.SequentialEvent(
    >>>             [
    >>>                 basic.SimpleEvent(random.uniform(0.5, 2))
    >>>                 for _ in range(random.randint(2, 5))
    >>>             ]
    >>>         )
    >>>         for _ in range(random.randint(1, 3))
    >>>     ]
    >>> )
    >>> DurationPrintConverter().convert(random_event)
    0: 1.182390506771032
    1.182390506771032: 1.6561757084885333
    2.8385662152595654: 1.558269840401042
    4.396836055660607: 1.5979384595498836
    5.994774515210491: 1.1502716523431056
    """

    @abc.abstractmethod
    def _convert_simple_event(
        self,
        event_to_convert: events.basic.SimpleEvent,
        absolute_entry_delay: constants.DurationType,
    ) -> typing.Sequence[typing.Any]:
        """Convert instance of :class:`mutwo.events.basic.SimpleEvent`."""

        raise NotImplementedError

    def _convert_simultaneous_event(
        self,
        simultaneous_event: events.basic.SimultaneousEvent,
        absolute_entry_delay: constants.DurationType,
    ) -> typing.Sequence[typing.Any]:
        """Convert instance of :class:`mutwo.events.basic.SimultaneousEvent`."""

        data_per_simple_event_list: list[tuple[typing.Any]] = []

        for event in simultaneous_event:
            data_per_simple_event_list.extend(
                self._convert_event(event, absolute_entry_delay)
            )
        return tuple(data_per_simple_event_list)

    def _convert_sequential_event(
        self,
        sequential_event: events.basic.SequentialEvent,
        absolute_entry_delay: constants.DurationType,
    ) -> typing.Sequence[typing.Any]:
        """Convert instance of :class:`mutwo.events.basic.SequentialEvent`."""

        data_per_simple_event_list: list[tuple[typing.Any]] = []
        for event_start, event in zip(
            sequential_event.absolute_time_tuple, sequential_event
        ):
            data_per_simple_event_list.extend(
                self._convert_event(event, event_start + absolute_entry_delay)
            )
        return tuple(data_per_simple_event_list)

    def _convert_event(
        self,
        event_to_convert: events.abc.Event,
        absolute_entry_delay: constants.DurationType,
    ) -> typing.Any:
        """Convert :class:`mutwo.events.abc.Event` of unknown type.

        The method calls different subroutines depending on whether
        the passed event either are instances from:

            1. :class:`mutwo.events.basic.SimpleEvent` or
            2. :class:`mutwo.events.basic.SequentialEvent` or
            3. :class:`mutwo.events.basic.SimultaneousEvent`.
        """

        if isinstance(event_to_convert, events.basic.SequentialEvent):
            return self._convert_sequential_event(
                event_to_convert, absolute_entry_delay
            )

        elif isinstance(
            event_to_convert,
            events.basic.SimultaneousEvent,
        ):
            return self._convert_simultaneous_event(
                event_to_convert, absolute_entry_delay
            )

        elif isinstance(
            event_to_convert,
            events.basic.SimpleEvent,
        ):
            return self._convert_simple_event(event_to_convert, absolute_entry_delay)

        else:
            message = (
                "Can't convert object '{}' of type '{}' with EventConverter.".format(
                    event_to_convert, type(event_to_convert)
                )
            )

            message += " Supported types only include all inherited classes "
            message += "from '{}'.".format(events.abc.Event)
            raise TypeError(message)


class SymmetricalEventConverter(EventConverter):
    """Abstract base class for Converter which handle mutwo events.

    This converter is a more specified version of the :class:`EventConverter`.
    It helps for building converters which aim to return mutwo events.
    """

    @abc.abstractmethod
    def _convert_simple_event(
        self,
        event_to_convert: events.basic.SimpleEvent,
        absolute_entry_delay: constants.DurationType,
    ) -> events.basic.SimpleEvent:
        """Convert instance of :class:`mutwo.events.basic.SimpleEvent`."""

        raise NotImplementedError

    def _convert_simultaneous_event(
        self,
        simultaneous_event: events.basic.SimultaneousEvent,
        absolute_entry_delay: constants.DurationType,
    ) -> events.basic.SimultaneousEvent:
        """Convert instance of :class:`mutwo.events.basic.SimultaneousEvent`."""

        converted_simultaneous_event: events.basic.SimultaneousEvent = (
            simultaneous_event.empty_copy()
        )

        for event in simultaneous_event:
            converted_simultaneous_event.append(
                self._convert_event(event, absolute_entry_delay)
            )
        return converted_simultaneous_event

    def _convert_sequential_event(
        self,
        sequential_event: events.basic.SequentialEvent,
        absolute_entry_delay: constants.DurationType,
    ) -> events.basic.SequentialEvent:
        """Convert instance of :class:`mutwo.events.basic.SequentialEvent`."""

        converted_sequential_event: events.basic.SequentialEvent = (
            sequential_event.empty_copy()
        )
        for event_start, event in zip(
            sequential_event.absolute_time_tuple, sequential_event
        ):
            converted_sequential_event.append(
                self._convert_event(event, event_start + absolute_entry_delay)
            )
        return converted_sequential_event

    def _convert_event(
        self,
        event_to_convert: events.abc.Event,
        absolute_entry_delay: constants.DurationType,
    ) -> events.abc.ComplexEvent[events.abc.Event]:
        return super()._convert_event(event_to_convert, absolute_entry_delay)



"""Classes and functions that generate data with the potential of artistic use.

The module is organised in different submodules where each submodule is named after the first known
person who introduced the respective algorithms. Unlike the :mod:`mutwo.converters` module the entered
data and the resulting data can be very different in type and form.

The term 'generators' simply labels the functionality of the module and shouldn't be confused with the Python
term for specific functions with the 'yield' keyword.
"""

from . import generic



"""Generic utility functions."""

import bisect
import copy
import functools
import importlib
import inspect
import itertools
import math
import operator
import pkgutil
import types
import typing
import warnings

from mutwo.core.utilities import constants


__all__ = (
    # "fetch_called_module"  # not for public use
    # "fetch_doc_string_from_core_module"  # not for public use
    # "import_all_submodules"  # not for public use
    "scale",
    "scale_sequence_to_sum",
    "accumulate_from_n",
    "accumulate_from_zero",
    "insert_next_to",
    "uniqify_sequence",
    "cyclic_permutations",
    # "import_module_if_dependencies_have_been_installed",  # not for public use
    "find_closest_index",
    "find_closest_item",
    # "class_name_to_object_name",  # not for public use
    "get_nested_item_from_index_sequence",
    "set_nested_item_from_index_sequence",
    "find_numbers_which_sums_up_to",
    # call_function_except_attribute_error  # not for public use
)


def fetch_called_module() -> types.ModuleType:
    """Fetch module from which the function got called.

    Raises error if function has not been called from a module.
    """

    frame = inspect.stack()[2]
    module = inspect.getmodule(frame[0])
    if module:
        return module
    else:
        raise NotImplementedError("Function hasn't been called from a module!")


def fetch_doc_string_from_core_module(module: typing.Optional[types.ModuleType] = None):
    """Fetch doc string for ext module __init__ file from core

    :param module: The mutwo.ext module.
    :type module: typing.Optional[types.ModuleType]
    """

    if not module:
        module = fetch_called_module()

    ext_module_name = ".".join(module.__name__.split(".")[2:])
    core_module = importlib.import_module(f"mutwo.core.{ext_module_name}")
    module.__doc__ = core_module.__doc__


def import_all_submodules(module: typing.Optional[types.ModuleType] = None):
    """Import all submodules from current module.

    :param module: The module which submodules shall be imported.
    :type module: typing.Optional[types.ModuleType]
    :return: None

    The method will import submodules in an order which is specified
    by the submodules name:

    1. It will first import all modules which are named "constants"
        (e.g. 'pitches_constants', 'basic_constants', etc.)

    2. Next it will import a module with the exact name "abc"
        (which is supposed to contain the public API for all other
        classes in the other submodules)

    3. Finally all other modules get imported.
    """

    def module_name_to_sort_index(module_name: str) -> int:
        if "constants" in module_name:
            return 0
        elif "abc" == module_name:
            return 1
        else:
            return 2

    if not module:
        module = fetch_called_module()

    module_name_list = sorted(
        list(map(operator.itemgetter(1), pkgutil.walk_packages(module.__path__))),
        key=module_name_to_sort_index,
    )
    for name in module_name_list:
        try:
            importlib.import_module(module.__name__ + "." + name)
        except ModuleNotFoundError:
            pass


def initialise_mutwo_ext_init():
    """Setup __init__.py file for mutwo.ext module"""

    module = fetch_called_module()
    fetch_doc_string_from_core_module(module)
    import_all_submodules(module)


def scale(
    value: constants.Real,
    old_min: constants.Real,
    old_max: constants.Real,
    new_min: constants.Real,
    new_max: constants.Real,
    translation_shape: constants.Real = 0,
) -> constants.Real:
    """Scale a value from one range to another range.

    :param value: The value that shall be scaled.
    :param old_min: The minima of the old range.
    :param old_max: The maxima of the old range.
    :param new_min: The minima of the new range.
    :param new_max: The maxima of the new range.
    :param translation_shape: 0 for a linear translation,
        values > 0 for a slower change at the beginning,
        values < 0 for a faster change at the beginning.

    The algorithmic to change the translation with the
    `translation_shape` has been copied from
    `expenvelope <https://git.sr.ht/~marcevanstein/expenvelope/tree/master/item/expenvelope/envelope_segment.py#L206>`_
    by M. Evanstein.

    **Example:**

    >>> from mutwo.core.utilities import tools
    >>> tools.scale(1, 0, 1, 0, 100)
    100
    >>> tools.scale(0.5, 0, 1, 0, 100)
    50
    >>> tools.scale(0.2, 0, 1, 0, 100)
    20
    >>> tools.scale(0.2, 0, 1, 0, 100, 1)
    12.885124808584155
    >>> tools.scale(0.2, 0, 1, 0, 100, -1)
    28.67637263023771
    """

    try:
        assert old_min <= value <= old_max
    except AssertionError:
        raise ValueError(
            f"Input value '{value}' has to be in the range of "
            f"(old_min = {old_min}, old_max = {old_max})."
        )

    percentage = (value - old_min) / (old_max - old_min)
    new_range = new_max - new_min
    if translation_shape:
        value = (new_range / ((math.exp(translation_shape)) - 1)) * (
            math.exp(translation_shape * percentage) - 1
        )
    else:
        value = new_range * percentage
    return value + new_min


def scale_sequence_to_sum(
    sequence_to_scale: typing.Sequence[constants.Real], sum_to_scale_to: constants.Real
) -> typing.Sequence[constants.Real]:
    """Scale numbers in a sequence so that the resulting sum fits to the given value.

    :param sequence_to_scale: The sequence filled with real numbers which sum should fit
        to the given `sum_to_scale_to` argument.
    :type sequence_to_scale: typing.Sequence[constants.Real]
    :param sum_to_scale_to: The resulting sum of the sequence.
    :type sum_to_scale_to: constants.Real

    **Example:**

    >>> from mutwo import utilities
    >>> sequence_to_scale = [1, 3, 2]
    >>> utilities.tools.scale_sequence_to_sum(sequence_to_scale, 3)
    [0.5, 1.5, 1]
    """

    if sequence_to_scale:
        current_sum = sum(sequence_to_scale)
        if current_sum:
            factor = sum_to_scale_to / current_sum
            scaled_sequence = map(lambda number: number * factor, sequence_to_scale)
        else:
            n_items_to_scale = len(sequence_to_scale)
            size_per_item = sum_to_scale_to / n_items_to_scale
            scaled_sequence = (size_per_item for _ in sequence_to_scale)
        return type(sequence_to_scale)(scaled_sequence)
    else:
        return copy.copy(sequence_to_scale)


def accumulate_from_n(
    iterable: typing.Iterable[constants.Real], n: constants.Real
) -> typing.Iterator:
    """Accumulates iterable starting with value n.

    :param iterable: The iterable which values shall be accumulated.
    :param n: The start number from which shall be accumulated.

    **Example:**

    >>> from mutwo.utilities import tools
    >>> tools.accumulate_from_n((4, 2, 3), 0)
    (0, 4, 6, 9)
    >>> tools.accumulate_from_n((4, 2, 3), 2)
    (2, 6, 8, 11)
    """
    return itertools.accumulate(itertools.chain((n,), iterable))


def accumulate_from_zero(iterable: typing.Iterable[constants.Real]) -> typing.Iterator:
    """Accumulates iterable starting from 0.

    :param iterable: The iterable which values shall be accumulated.

    **Example:**

    >>> from mutwo.utilities import tools
    >>> tools.accumulate_from_zero((4, 2, 3), 0)
    (0, 4, 6, 9)
    """
    return accumulate_from_n(iterable, 0)


def insert_next_to(
    mutable_sequence: typing.MutableSequence,
    item_to_find: typing.Any,
    distance: int,
    item_to_insert: typing.Any,
):
    """Insert an item into a list relative to the first item equal to a certain value."""

    index = list(mutable_sequence).index(item_to_find)
    if distance == 0:
        mutable_sequence[index] = item_to_insert
    else:
        real_distance = distance + 1 if distance < 0 else distance
        mutable_sequence.insert(index + real_distance, item_to_insert)


T = typing.TypeVar("T", bound=constants.Real)


def find_closest_index(
    item: constants.Real,
    sequence: typing.Sequence,
    key: typing.Callable[[typing.Any], T] = lambda item: item,
) -> int:
    """Return index of element in ``data`` with smallest difference to ``item``.

    :param item: The item from which the closest item shall be found.
    :param sequence: The data to which the closest item shall be found.

    **Example:**

    >>> from mutwo.utilities import tools
    >>> tools.find_closest_index(2, (1, 4, 5))
    0
    >>> tools.find_closest_index(127, (100, 4, 300, 53, 129))
    4
    >>> tools.find_closest_index(127, (('hi', 100), ('hey', 4), ('hello', 300)), key=lambda item: item[1])
    0
    """

    research_data = tuple(map(key, sequence))
    sorted_research_data = sorted(research_data)

    solution = bisect.bisect_left(sorted_research_data, item)
    # make type ignore because data has been converted to tuple (which is sizeable)
    if solution == len(sequence):
        index = solution - 1

    elif solution == 0:
        index = solution

    else:
        indices = (solution, solution - 1)
        differences = tuple(abs(-sorted_research_data[n] + item) for n in indices)
        index = indices[differences.index(min(differences))]  # type: ignore

    return research_data.index(sorted_research_data[index])


def find_closest_item(
    item: constants.Real,
    sequence: typing.Sequence,
    key: typing.Callable[[typing.Any], T] = lambda item: item,
) -> T:
    """Return element in ``data`` with smallest difference to ``item``.

    :param item: The item from which the closest item shall be found.
    :param sequence: The data to which the closest item shall be found.
    :return: The closest number to ``item`` in ``data``.

    **Example:**

    >>> from mutwo.utilities import tools
    >>> tools.find_closest_item(2, (1, 4, 5))
    1
    >>> tools.find_closest_item(127, (100, 4, 300, 53, 129))
    129
    >>> tools.find_closest_item(127, (('hi', 100), ('hey', 4), ('hello', 300)), key=lambda item: item[1])
    ('hi', 100)
    """
    return sequence[find_closest_index(item, sequence, key=key)]


def import_module_if_dependencies_have_been_installed(
    module: str, dependencies: tuple[str, ...], import_class: bool = False
) -> None:
    for dependency in dependencies:
        if not importlib.util.find_spec(dependency):
            message = (
                "Can't load module '{0}'. Install dependency '{1}' if you want to use"
                " '{0}'.".format(module, dependency)
            )
            warnings.warn(message)
            return

    if import_class:
        class_name = module.split(".")[-1]
        return getattr(importlib.import_module(module), class_name)

    else:
        importlib.import_module(module)


def uniqify_sequence(
    sequence: typing.Sequence,
    sort_key: typing.Callable[[typing.Any], constants.Real] = None,
    group_by_key: typing.Callable[[typing.Any], typing.Any] = None,
) -> typing.Iterable:
    """Not-Order preserving function to uniqify any iterable with non-hashable objects.

    :param sequence: The iterable which items shall be uniqified.
    :return: Return uniqified version of the entered iterable.
        The function will try to return the same type of the passed
        iterable. If Python raises an error during initialisation of
        the original iterable type, the function will simply return
        a tuple.

    **Example:**

    >>> from mutwo.parameters import pitches
    >>> from mutwo.utilities import tools
    >>> tools.uniqify_sequence([pitches.WesternPitch(pitch_name) for pitch_name in 'c d e c d e e f a c a'.split(' ')])
    [WesternPitch(c4),
    WesternPitch(d4),
    WesternPitch(e4),
    WesternPitch(f4),
    WesternPitch(a4)]
    """

    sorted_iterable = sorted(sequence, key=sort_key)
    result = (key for key, _ in itertools.groupby(sorted_iterable, key=group_by_key))

    try:
        return type(sequence)(result)  # type: ignore

    except Exception:
        return tuple(result)


def cyclic_permutations(sequence: typing.Sequence[typing.Any]) -> typing.Generator:
    """Cyclic permutation of an iterable. Return a generator object.

    :param sequence: The sequence from which cyclic permutations shall be generated.

    **Example:**

    >>> from mutwo.utilities import tools
    >>> permutations = tools.cyclic_permutations((1, 2, 3, 4))
    >>> next(permutations)
    (2, 3, 4, 1)
    >>> next(permutations)
    (3, 4, 1, 2)

    `Adapted function from the reply of Paritosh Singh
    <https://stackoverflow.com/questions/56171246/cyclic-permutation-operators-in-python/56171531>`_
    """

    def reorder_from_index(index: int, sequence: tuple) -> tuple:
        return sequence[index:] + sequence[:index]

    return (
        functools.partial(reorder_from_index, i)(sequence) for i in range(len(sequence))
    )


def camel_case_to_snake_case(camel_case_string: str) -> str:
    """Transform camel case formatted string to snake case.

    :param camel_case_string: String which is formatted using
        camel case (no whitespace, but upper letters at
        new word start).
    :return: string formatted using snake case

    **Example:** MyClassName -> my_class_name
    """

    character_list = []

    is_first = True
    for character in camel_case_string:
        if character.isupper():
            character = character.lower()
            if not is_first:
                character = "_{}".format(character)

        if is_first:
            is_first = False

        character_list.append(character)

    return "".join(character_list)


def get_nested_item_from_index_sequence(
    index_sequence: typing.Sequence[int], sequence: typing.Sequence
) -> typing.Any:
    """Get item in nested Sequence.

    :param index_sequence: The indices of the nested item.
    :type index_sequence: typing.Sequence[int]
    :param sequence: A nested sequence.
    :type sequence: typing.Sequence[typing.Any]


    **Example:**

    >>> from mutwo.utilities import tools
    >>> nested_sequence = (1, 2, (4, (5, 1), (9, (3,))))
    >>> tools.get_nested_item_from_index_sequence((2, 2, 0), nested_sequence)
    9
    >>> nested_sequence[2][2][0]  # is equal
    9
    """

    for index in index_sequence:
        sequence = sequence[index]
    return sequence


def set_nested_item_from_index_sequence(
    index_sequence: typing.Sequence[int],
    sequence: typing.MutableSequence,
    item: typing.Any,
) -> None:
    """Set item in nested Sequence.

    :param index_sequence: The indices of the nested item which shall be set.
    :type index_sequence: typing.Sequence[int]
    :param sequence: A nested sequence.
    :type sequence: typing.MutableSequence[typing.Any]
    :param item: The new item value.
    :type item: typing.Any

    **Example:**

    >>> from mutwo.utilities import tools
    >>> nested_sequence = [1, 2, [4, [5, 1], [9, [3]]]]]
    >>> tools.set_nested_item_from_index_sequence((2, 2, 0), nested_sequence, 100)
    >>> nested_sequence[2][2][0] = 100  # is equal
    """

    n_indices = len(index_sequence)
    for nth_index, index in enumerate(index_sequence):
        if n_indices == nth_index + 1:
            sequence.__setitem__(index, item)
        else:
            sequence = sequence[index]


def round_floats(number_to_round: constants.Real, n_digits: int) -> constants.Real:
    """Round number if it is an instance of float, otherwise unaltered number.

    :param number_to_round: The number which shall be rounded.
    :type number_to_round: constants.Real
    :param n_digits: How many digits shall the number be rounded.
    :type n_digits: int
    """

    if isinstance(number_to_round, float):
        return round(number_to_round, n_digits)
    else:
        return number_to_round


def find_numbers_which_sums_up_to(
    given_sum: float,
    number_to_choose_from_sequence: typing.Optional[typing.Sequence[float]] = None,
    n_items_to_sum_up: typing.Optional[set[int]] = None,
) -> tuple[tuple[float, ...], ...]:
    """Find all combinations of numbers which sum is equal to the given sum.

    :param given_sum: The target sum for which different combinations shall
        be searched.
    :type given_sum: float
    :param number_to_choose_from_sequence: A sequence of numbers which shall be
        tried to combine to result in the :attr:`given_sum`. If the user
        doesn't specify this argument mutwo will use all natural numbers
        equal or smaller than the :attr:`given_sum`.
    :type number_to_choose_from_sequence: typing.Optional[typing.Sequence[float]]
    :param n_items_to_sum_up: How many numbers can be combined to result
        in the :attr:`given_sum`. If the user doesn't specify this argument
        mutwo will use all natural numbers equal or smaller than the
        :attr:`given_sum`.
    :type n_items_to_sum_up: typing.Optional[set[int]]

    **Example:**

    >>> from mutwo.utilities import tools
    >>> tools.find_numbers_which_sums_up_to(4)
    ((4,), (1, 3), (2, 2), (1, 1, 2), (1, 1, 1, 1))
    """

    if not number_to_choose_from_sequence:
        number_to_choose_from_sequence = tuple(range(1, int(given_sum) + 1))

    if not n_items_to_sum_up:
        n_items_to_sum_up = set(range(1, given_sum + 1))

    numbers = []
    for n_items in n_items_to_sum_up:
        numbers.extend(
            [
                pair
                for pair in itertools.combinations_with_replacement(
                    number_to_choose_from_sequence, n_items
                )
                if sum(pair) == given_sum
            ]
        )
    return tuple(numbers)


def call_function_except_attribute_error(
    function: typing.Callable[[typing.Any], typing.Any],
    argument: typing.Any,
    exception_value: typing.Any,
) -> typing.Any:
    """Run a function with argument as input

    :param function: The function to be called.
    :param argument: The argument with which the function shall be called.
    :param exception_value: The alternative value if the function call raises an
        `AttributeError`.
    :return: Return :obj:`exception_value` in case an attribute error occurs.
        In case the function call is successful the function return value will
        be returned.
    """

    try:
        value = function(argument)
    except AttributeError:
        value = exception_value

    return value



"""Generic algorithms which are designed by mutwo authors."""

import numbers
import typing

import expenvelope  # type: ignore
import numpy as np  # type: ignore

from mutwo.core import utilities

__all__ = ("DynamicChoice",)


class DynamicChoice(object):
    """Weighted random choices with dynamically changing weights.

    :param value_sequence: The items to choose from.
    :type value_sequence: typing.Sequence[typing.Any]
    :param curve_sequence: The dynamically changing weight for each value.
    :type curve_sequence: typing.Sequence[expenvelope.Envelope]
    :param random_seed: The seed which shall be set at class initialisation.
    :type random_seed: int

    **Example:**

    >>> import expenvelope
    >>> from mutwo.generators import generic
    >>> dynamic_choice = generic.DynamicChoice(
    >>>    [0, 1, 2],
    >>>    [
    >>>        expenvelope.Envelope.from_points((0, 0), (0.5, 1), (1, 0)),
    >>>        expenvelope.Envelope.from_points((0, 0.5), (0.5, 0), (1, 0.5)),
    >>>        expenvelope.Envelope.from_points((0, 0.5), (1, 1)),
    >>>    ],
    >>> )
    >>> dynamic_choice.gamble_at(0.3)
    2
    >>> dynamic_choice.gamble_at(0.3)
    2
    >>> dynamic_choice.gamble_at(0.3)
    0
    """

    def __init__(
        self,
        value_sequence: typing.Sequence[typing.Any],
        curve_sequence: typing.Sequence[expenvelope.Envelope],
        random_seed: int = 100,
    ):

        assert len(value_sequence) == len(curve_sequence)

        self._value_sequence = value_sequence
        self._curve_sequence = curve_sequence
        self._random = np.random.default_rng(random_seed)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._value_sequence})"

    def items(self) -> tuple[tuple[typing.Any, expenvelope.Envelope]]:
        return tuple(zip(self._value_sequence, self._curve_sequence))

    def gamble_at(self, time: numbers.Real) -> typing.Any:
        """Return value at requested time.

        :param time: At which position on the x-Axis shall be gambled.
        :type time: numbers.Real
        :return: The chosen value.
        """
        weight_list = [curve.value_at(time) for curve in self._curve_sequence]
        return self._random.choice(
            self._value_sequence, p=utilities.tools.scale_sequence_to_sum(weight_list, 1)
        )



"""Apply tempo curve on any :class:`~mutwo.events.abc.Event` and convert :class:`~mutwo.parameters.tempos.TempoPoint` to beat-length-in-seconds.

"""

import typing
import warnings

import expenvelope  # type: ignore

from mutwo.core import converters, events, parameters
from mutwo.core.utilities import constants


TempoEvents = expenvelope.Envelope
TempoPoint = typing.Union[parameters.tempos.TempoPoint, constants.Real]


__all__ = ("TempoPointConverter", "TempoConverter",)


class TempoPointConverter(converters.abc.Converter):
    """Convert a :class:`~mutwo.parameters.tempos.TempoPoint` with BPM to beat-length-in-seconds.

    A :class:`TempoPoint` is defined as an object that has a particular tempo in
    beats per seconds (BPM) and a reference value (1 for a quarter note, 4
    for a whole note, etc.). Besides elaborate :class:`mutwo.parameters.tempos.TempoPoint`
    objects, any number can also be interpreted as a `TempoPoint`. In this case
    the number simply represents the BPM number and the reference will be set to 1.
    The returned beat-length-in-seconds always indicates the length for one quarter
    note.

    **Example:**

    >>> from mutwo.converters import symmetrical
    >>> tempo_point_converter = symmetrical.tempos.TempoPointConverter()
    """

    @staticmethod
    def _beats_per_minute_to_seconds_per_beat(
        beats_per_minute: constants.Real,
    ) -> float:
        return float(60 / beats_per_minute)

    @staticmethod
    def _extract_beats_per_minute_and_reference_from_tempo_point(
        tempo_point: TempoPoint,
    ) -> tuple[constants.Real, constants.Real]:
        try:
            beats_per_minute = tempo_point.tempo_in_beats_per_minute  # type: ignore
        except AttributeError:
            beats_per_minute = float(tempo_point)  # type: ignore

        try:
            reference = tempo_point.reference  # type: ignore
        except AttributeError:
            message = (
                "Tempo point {} of type {} doesn't know attribute 'reference'.".format(
                    tempo_point, type(tempo_point)
                )
            )
            message += " Therefore reference has been set to 1."
            warnings.warn(message)
            reference = 1

        return beats_per_minute, reference

    def convert(self, tempo_point_to_convert: TempoPoint) -> float:
        """Converts a :class:`TempoPoint` to beat-length-in-seconds.

        :param tempo_point_to_convert: A tempo point defines the active tempo
            from which the beat-length-in-seconds shall be calculated. The argument
            can either be any number (which will be interpreted as beats per
            minute [BPM]) or a ``mutwo.parameters.tempos.TempoPoint`` object.
        :return: The duration of one beat in seconds within the passed tempo.

        **Example:**

        >>> from mutwo.converters import symmetrical
        >>> converter = symmetrical.tempos.TempoPointConverter()
        >>> converter.convert(60)  # one beat in tempo 60 bpm takes 1 second
        1
        >>> converter.convert(120)  # one beat in tempo 120 bpm takes 0.5 second
        0.5
        """

        (
            beats_per_minute,
            reference,
        ) = self._extract_beats_per_minute_and_reference_from_tempo_point(
            tempo_point_to_convert
        )
        return (
            TempoPointConverter._beats_per_minute_to_seconds_per_beat(beats_per_minute)
            / reference
        )


class TempoConverter(converters.abc.EventConverter):
    """Class for applying tempo curves on mutwo events.

    :param tempo_envelope: The tempo curve that shall be applied on the
        mutwo events. This is expected to be a :class:`expenvelope.Envelope`
        which levels are filled with numbers that will be interpreted as BPM
        [beats per minute]) or with :class:`mutwo.parameters.tempos.TempoPoint`
        objects.

    **Example:**

    >>> import expenvelope
    >>> from mutwo.converters import symmetrical
    >>> from mutwo.parameters import tempos
    >>> tempo_envelope = expenvelope.Envelope.from_levels_and_durations(
    >>>     levels=[tempos.TempoPoint(60), 60, 30, 50],
    >>>     durations=[3, 0, 2],
    >>> )
    >>> my_tempo_converter = symmetrical.tempos.TempoConverter(tempo_envelope)
    """

    _tempo_point_converter = TempoPointConverter()

    def __init__(self, tempo_envelope: expenvelope.Envelope):
        self._envelope = TempoConverter._tempo_envelope_to_beat_length_in_seconds_envelope(
            tempo_envelope
        )

    # ###################################################################### #
    #                          static methods                                #
    # ###################################################################### #

    @staticmethod
    def _tempo_envelope_to_beat_length_in_seconds_envelope(
        tempo_envelope: expenvelope.Envelope,
    ) -> expenvelope.Envelope:
        """Convert bpm / TempoPoint based env to beat-length-in-seconds env."""

        level_list: list[float] = []
        for tempo_point in tempo_envelope.levels:
            beat_length_in_seconds = TempoConverter._tempo_point_converter.convert(
                tempo_point
            )
            level_list.append(beat_length_in_seconds)

        return expenvelope.Envelope.from_levels_and_durations(
            level_list, tempo_envelope.durations, tempo_envelope.curve_shapes
        )

    # ###################################################################### #
    #                         private methods                                #
    # ###################################################################### #

    def _convert_simple_event(
        self,
        simple_event: events.basic.SimpleEvent,
        absolute_entry_delay: constants.DurationType,
    ) -> tuple[typing.Any, ...]:
        simple_event.duration = self._envelope.integrate_interval(
            absolute_entry_delay, simple_event.duration + absolute_entry_delay
        )
        return tuple([])

    # ###################################################################### #
    #               public methods for interaction with the user             #
    # ###################################################################### #

    def convert(self, event_to_convert: events.abc.Event) -> events.abc.Event:
        """Apply tempo curve of the converter to the entered event.

        The method doesn't change the original event, but returns a copied
        version with different values for its duration attributes depending
        on the tempo curve.

        :param event_to_convert: The event to convert. Can be any object
            that inherits from ``mutwo.events.abc.Event``. If the event that
            shall be converted is longer than the tempo curve of the
            ``TempoConverter``, then the last tempo of the curve will be hold.
        :return: A new ``Event`` object which duration property has been adapted
            by the tempo curve of the ``TempoConverter``.

        **Example:**

        >>> import expenvelope
        >>> from mutwo.events import basic
        >>> from mutwo.parameters import tempos
        >>> from mutwo.converters import symmetrical
        >>> tempo_envelope = expenvelope.Envelope.from_levels_and_durations(
        >>>     levels=[tempos.TempoPoint(60), 60, 120, 120],
        >>>     durations=[3, 2, 5],
        >>> )
        >>> my_tempo_converter = symmetrical.tempos.TempoConverter(tempo_envelope)
        >>> my_events = basic.SequentialEvent([basic.SimpleEvent(d) for d in (3, 2, 5)])
        >>> my_tempo_converter.convert(my_events)
        SequentialEvent([SimpleEvent(duration = 3.0), SimpleEvent(duration = 1.5), SimpleEvent(duration = 2.5)])
        """
        copied_event_to_convert = event_to_convert.destructive_copy()
        self._convert_event(copied_event_to_convert, 0)
        return copied_event_to_convert



import setuptools  # type: ignore


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

extras_require = {"testing": ["nose", "pillow>=8.2.0, <9.0.0", "mutwo.ext-music>=0.1.0, <1.0.0"]}

setuptools.setup(
    name="mutwo",
    version="0.48.1",
    license="GPL",
    description="event based framework for generative art",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Tim Pauli,  Levin Eric Zimmermann",
    author_email="tim.pauli@folkwang-uni.de, levin.eric.zimmermann@posteo.eu",
    url="https://github.com/mutwo-org/mutwo",
    project_urls={"Documentation": "https://mutwo.readthedocs.io/en/latest/"},
    packages=[
        package for package in setuptools.find_packages() if package[:5] != "tests"
    ],
    setup_requires=[],
    install_requires=[
        "dill>=0.3.4, <1.0.0",
        "expenvelope>=0.6.5, <1.0.0",
        "primesieve>=2.0.0, <3.0.0",
        "numpy>=1.18, <2.00",
        "scipy>=1.4.1, <2.0.0",
        "python-ranges>=0.2.0, <1.0.0",
    ],
    extras_require=extras_require,
    python_requires=">=3.9, <4",
)



"""Prime number calculations.

The functions "factorise" and "factors" are copied from the `pyprimes
library <https://github.com/uzumaxy/pyprimes/blob/master/src/pyprimes/factors.py>`_.
In mutwo the function :func:`pyprimes.primes` has been replaced by
:class:`primesieve.Iterator()`, which improves speed by a factor of 10.
"""

import typing

import primesieve  # type: ignore

__all__ = ("factorise", "factors", "is_prime")


class PrimeGenerator(object):
    """Generator that yields the rising series of primes starting from 2."""

    def __init__(self):
        self.it = primesieve.Iterator()

    def __next__(self) -> int:
        return self.it.next_prime()

    def __iter__(self) -> typing.Iterator:
        return self


def factorise(n: int) -> list[int]:
    """factorise(integer) -> [list of factors]

    :param n: The number which shall be factorised.
    :return: Returns a list of the (mostly) prime factors of integer n. For negative
        integers, -1 is included as a factor. If n is 0, 1 or -1, [n] is
        returned as the only factor. Otherwise all the factors will be prime.

    **Example:**

    >>> factorise(-693)
    [-1, 3, 3, 7, 11]
    >>> factorise(55614)
    [2, 3, 13, 23, 31]
    """
    result = []
    for p, count in factors(n):
        result.extend([p] * count)
    return result


def factors(n: int):
    """factors(integer) -> yield factors of integer lazily

    Yields tuples of (factor, count) where each factor is unique and usually
    prime, and count is an integer 1 or larger.
    The factors are prime, except under the following circumstances: if the
    argument n is negative, -1 is included as a factor; if n is 0 or 1, it
    is given as the only factor. For all other integer n, all of the factors
    returned are prime.

    **Example:**

    >>> list(factors(3*7*7*7*11))
    [(3, 1), (7, 3), (11, 1)]
    """
    if n in (0, 1, -1):
        yield (n, 1)
        return
    elif n < 0:
        yield (-1, 1)
        n = -n
    assert n >= 2
    for p in PrimeGenerator():
        if p * p > n:
            break
        count = 0
        while n % p == 0:
            count += 1
            n //= p
        if count:
            yield (p, count)
    if n != 1:
        yield (n, 1)


def is_prime(n: int) -> bool:
    """Test if number is prime or not.

    :param n: The number which shall be tested.
    :return: True if number is prime and False if number isn't a Prime.

    (`has been copied from here
    <https://www.geeksforgeeks.org/python-program-to-check-
    whether-a-number-is-prime-or-not/>`_)
    """

    # Corner cases
    if n <= 1:
        return False
    if n <= 3:
        return True

    # This is checked so that we can skip
    # middle five numbers in below loop
    if n % 2 == 0 or n % 3 == 0:
        return False

    i = 5
    while i * i <= n:
        if n % i == 0 or n % (i + 2) == 0:
            return False
        i = i + 6

    return True



"""This module adds several constants that are used for the different pitch classes.

There are basically two types of constants:

1. Default values for initialisation method argument of the different pitch classes.
2. Values that are defined by general (cultural, physical, technical) standards
"""

import math

try:
    import quicktions as fractions  # type: ignore
except ImportError:
    import fractions  # type: ignore

from mutwo.core.parameters import commas

CENT_CALCULATION_CONSTANT = 1200 / (math.log10(2))
"""constant used for cent calculation in mutwo.parameters.abc.Pitch"""

DIATONIC_PITCH_NAME_TO_PITCH_CLASS_DICT = {
    diatonic_pitch_name: pitch_class
    for diatonic_pitch_name, pitch_class in zip(
        "c d e f g a b".split(" "), (0, 2, 4, 5, 7, 9, 11)
    )
}
"""Mapping of diatonic pitch name to pitch class for the `WesternPitch` class.
Mutwo uses a chromatic scale where a change of the number 1 is one half tone."""

ASCENDING_DIATONIC_PITCH_NAME_TUPLE = tuple(
    sorted(
        DIATONIC_PITCH_NAME_TO_PITCH_CLASS_DICT.keys(),
        key=lambda diatonic_pitch_name: DIATONIC_PITCH_NAME_TO_PITCH_CLASS_DICT[
            diatonic_pitch_name
        ],
    )
)
"""Tuple with diatonic pitch names in ascending order."""

ACCIDENTAL_NAME_TO_PITCH_CLASS_MODIFICATION_DICT = {
    # multiply with 2 because the difference of "1" in pitch
    # class is defined as one chromatic step (see class
    # definition of WesternPitch)
    accidental_name: accidental_value * 2
    for accidental_name, accidental_value in {
        # double sharp / double flat
        "ff": -fractions.Fraction(1, 1),
        "ss": fractions.Fraction(1, 1),
        # eleven twelfth-tone
        "etf": -fractions.Fraction(11, 12),
        "ets": fractions.Fraction(11, 12),
        # seven eigth-tone
        "sef": -fractions.Fraction(7, 8),
        "ses": fractions.Fraction(7, 8),
        # two third-tone
        "trf": -fractions.Fraction(2, 3),
        "trs": fractions.Fraction(2, 3),
        # three quarter-tone
        "tqf": -fractions.Fraction(3, 4),
        "tqs": fractions.Fraction(3, 4),
        # seven sixth-tone
        "sxf": -fractions.Fraction(7, 6),
        "sxs": fractions.Fraction(7, 6),
        # nine eight-tone
        "nef": -fractions.Fraction(9, 8),
        "nes": fractions.Fraction(9, 8),
        # seven twelfth-tone
        "stf": -fractions.Fraction(7, 12),
        "sts": fractions.Fraction(7, 12),
        # ordinary sharp / flat
        "f": -fractions.Fraction(1, 2),
        "s": fractions.Fraction(1, 2),
        # five twelfth-tone
        "ftf": -fractions.Fraction(5, 12),
        "fts": fractions.Fraction(5, 12),
        # three eigth-tone
        "tef": -fractions.Fraction(3, 8),
        "tes": fractions.Fraction(3, 8),
        # one third-tone (use "r" to avoid conufsion with twelfth-tone)
        "rf": -fractions.Fraction(1, 3),
        "rs": fractions.Fraction(1, 3),
        # one quarter-tone
        "qf": -fractions.Fraction(1, 4),
        "qs": fractions.Fraction(1, 4),
        # one sixth-tone (use x to avoid confusion with double-sharp ss)
        "xf": -fractions.Fraction(1, 6),
        "xs": fractions.Fraction(1, 6),
        # one eigth-tone
        "ef": -fractions.Fraction(1, 8),
        "es": fractions.Fraction(1, 8),
        # one twelfth-tone
        "tf": -fractions.Fraction(1, 12),
        "ts": fractions.Fraction(1, 12),
        # no accidental / empty string
        "": fractions.Fraction(0, 1),
    }.items()
}
"""Mapping of accidental name to pitch class modification for the
`WesternPitch` class. When adding new accidentals or changing
accidental names only this constant has to be changed."""

PITCH_CLASS_MODIFICATION_TO_ACCIDENTAL_NAME_DICT = {
    accidental_value: accidental_name
    for accidental_name, accidental_value in ACCIDENTAL_NAME_TO_PITCH_CLASS_MODIFICATION_DICT.items()
}
"""Mapping of pitch class modifications name accidental name for the
`WesternPitch` class. This global variable is defined in reference to
``ACCIDENTAL_NAME_TO_PITCH_CLASS_MODIFICATION``."""


# is used in mutwo.parameters.pitches.JustIntonationPitch
DIATONIC_PITCH_NAME_CYCLE_OF_FIFTH_TUPLE = tuple("f c g d a e b".split(" "))
"""Diatonic pitch names sorted by cycle of fifths."""

DEFAULT_CONCERT_PITCH = 440
"""The default concert pitch in frequency."""

DEFAULT_CONCERT_PITCH_OCTAVE_FOR_WESTERN_PITCH = 4
"""The default value of ``concert_pitch_octave_for_western_pitch`` for
the `WesternPitch` class. It sets octave of default concert pitch for
western pitch to one line (') (so that the resulting concert pitch is
a' or in midi nomenclature a4)"""

DEFAULT_CONCERT_PITCH_PITCH_CLASS_FOR_WESTERN_PITCH = 9
"""The default value of``concert_pitch_concert_pitch_class_for_western_pitch``
for the `WesternPitch` class. It sets pitch class 'a' as the default pitch class
for the concert pitch of the WesternPitch class."""

MIDI_PITCH_FREQUENCY_TUPLE = (
    8.175798915643705,
    8.66195721802725,
    9.177023997418985,
    9.722718241315027,
    10.300861153527185,
    10.91338223228137,
    11.562325709738575,
    12.249857374429666,
    12.978271799373287,
    13.750000000000,
    14.567617547440312,
    15.433853164253883,
    16.35159783128741,
    17.3239144360545,
    18.35404799483797,
    19.445436482630054,
    20.60172230705437,
    21.82676446456274,
    23.12465141947715,
    24.499714748859333,
    25.956543598746574,
    27.500000000000004,
    29.135235094880624,
    30.867706328507765,
    32.70319566257482,
    34.647828872109,
    36.70809598967594,
    38.89087296526011,
    41.20344461410874,
    43.65352892912548,
    46.2493028389543,
    48.999429497718666,
    51.91308719749315,
    55.00000000000001,
    58.27047018976125,
    61.73541265701553,
    65.40639132514964,
    69.295657744218,
    73.41619197935188,
    77.78174593052022,
    82.40688922821748,
    87.30705785825096,
    92.4986056779086,
    97.99885899543733,
    103.8261743949863,
    110.00000000000001,
    116.5409403795225,
    123.47082531403106,
    130.81278265029928,
    138.591315488436,
    146.83238395870376,
    155.56349186104043,
    164.81377845643496,
    174.6141157165019,
    184.9972113558172,
    195.99771799087466,
    207.6523487899726,
    220.00000000000003,
    233.081880759045,
    246.94165062806212,
    261.62556530059857,
    277.182630976872,
    293.6647679174075,
    311.12698372208087,
    329.6275569128699,
    349.2282314330038,
    369.9944227116344,
    391.9954359817493,
    415.3046975799452,
    440.00000000000006,
    466.16376151809,
    493.88330125612424,
    523.2511306011971,
    554.365261953744,
    587.329535834815,
    622.2539674441617,
    659.2551138257398,
    698.4564628660077,
    739.9888454232688,
    783.9908719634986,
    830.6093951598904,
    880.0000000000001,
    932.32752303618,
    987.7666025122485,
    1046.5022612023943,
    1108.730523907488,
    1174.65907166963,
    1244.5079348883235,
    1318.5102276514797,
    1396.9129257320153,
    1479.9776908465376,
    1567.9817439269973,
    1661.2187903197807,
    1760.0000000000002,
    1864.65504607236,
    1975.533205024497,
    2093.0045224047885,
    2217.461047814976,
    2349.31814333926,
    2489.015869776647,
    2637.0204553029594,
    2793.8258514640306,
    2959.955381693075,
    3135.9634878539946,
    3322.4375806395615,
    3520.0000000000005,
    3729.31009214472,
    3951.066410048994,
    4186.009044809577,
    4434.922095629952,
    4698.63628667852,
    4978.031739553294,
    5274.040910605919,
    5587.651702928061,
    5919.91076338615,
    6271.926975707989,
    6644.875161279123,
    7040.000000000001,
    7458.62018428944,
    7902.132820097988,
    8372.018089619154,
    8869.844191259905,
    9397.27257335704,
    9956.063479106588,
    10548.081821211837,
    11175.303405856122,
    11839.8215267723,
)
"""A tuple that contains the frequency of each midi pitch (from 1 to 127)."""

MIDI_PITCH_NUMBER_TUPLE = tuple(range(127))
"""A tuple that contains all available midi pitch numbers."""

# tuning commas are organised, so that the relevant prime number
# is always in the numerator of the ratio!
DEFAULT_PRIME_TO_COMMA_DICT = {
    # syntonic comma
    5: commas.Comma(fractions.Fraction(80, 81)),
    # septimal comma
    7: commas.Comma(fractions.Fraction(63, 64)),
    # undecimal quartertone
    11: commas.Comma(fractions.Fraction(33, 32)),
    # tridecimal thirdtone
    13: commas.Comma(fractions.Fraction(26, 27)),
    # 17-limit schisma
    17: commas.Comma(fractions.Fraction(2176, 2187)),
    # 19-limit schisma
    19: commas.Comma(fractions.Fraction(513, 512)),
    # 23-limit comma
    23: commas.Comma(fractions.Fraction(736, 729)),
    # 29-limit sixthtone
    29: commas.Comma(fractions.Fraction(261, 256)),
    # 31-limit quartertone
    31: commas.Comma(fractions.Fraction(31, 32)),
    # 37-limit quartertone
    37: commas.Comma(fractions.Fraction(37, 36)),
    # 41-limit comma
    41: commas.Comma(fractions.Fraction(82, 81)),
    # 43-limit comma
    43: commas.Comma(fractions.Fraction(129, 128)),
    # 47-limit quartertone
    47: commas.Comma(fractions.Fraction(752, 729)),
}
"""Standard commas defined by the `Helmholtz-Ellis JI Pitch Notation <https://marsbat.space/pdfs/notation.pdf>`_."""

PITCH_ENVELOPE_REFERENCE_FREQUENCY = 100
"""Reference frequency for internal calculation in
:class:`mutwo.core.parameters.abc.Pitch.PitchEnvelope`. Exact
number doesn't really matter, it only has to keep consistent."""

DEFAULT_PITCH_ENVELOPE_PARAMETER_NAME = "pitch"
"""Default property parameter name for events in
:class:`mutwo.core.parameters.abc.Pitch.PitchEnvelope`."""

DEFAULT_PITCH_INTERVAL_ENVELOPE_PARAMETER_NAME = "pitch_interval"
"""Default property parameter name for events in
:class:`mutwo.core.parameters.abc.Pitch.PitchIntervalEnvelope`."""
